"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Scope = exports.Definition = exports.Usage = exports.Locals = void 0;
const lsp = __importStar(require("vscode-languageserver"));
const common_1 = require("../common");
const languages_1 = __importDefault(require("../languages"));
class Locals {
    constructor(document, root) {
        this.document = document;
        this.root = root;
    }
    static create(document, trees) {
        const root = new Scope(lsp.Range.create(0, 0, document.lineCount, 0), true);
        const tree = trees.getParseTree(document);
        if (!tree) {
            return new Locals(document, root);
        }
        const all = [];
        const query = languages_1.default.getQuery(document.languageId, 'locals');
        const captures = query.captures(tree.rootNode).sort(this._compareCaptures);
        // Find all scopes and merge some. The challange is that function-bodies "see" their
        // arguments but function-block-nodes and argument-list-nodes are usually siblings
        const scopeCaptures = captures.filter(capture => capture.name.startsWith('scope'));
        for (let i = 0; i < scopeCaptures.length; i++) {
            const capture = scopeCaptures[i];
            const range = common_1.asLspRange(capture.node);
            if (capture.name.endsWith('.merge')) {
                all[all.length - 1].range.end = range.end;
            }
            else {
                all.push(new Scope(range, capture.name.endsWith('.exports')));
            }
        }
        // Find all definitions and usages and mix them with scopes
        this._fillInDefinitionsAndUsages(all, captures);
        //
        this._constructTree(root, all);
        const info = new Locals(document, root);
        // info.debugPrint();
        return info;
    }
    static _fillInDefinitionsAndUsages(bucket, captures) {
        for (const capture of captures) {
            if (capture.name.startsWith('local')) {
                bucket.push(new Definition(capture.node.text, common_1.asLspRange(capture.node)));
            }
            else if (capture.name.startsWith('usage')) {
                bucket.push(new Usage(capture.node.text, common_1.asLspRange(capture.node), capture.name.endsWith('.void')));
            }
        }
    }
    static _constructTree(root, nodes) {
        const stack = [];
        for (const thing of nodes.sort(this._compareByRange)) {
            while (true) {
                let parent = stack.pop() ?? root;
                if (common_1.containsRange(parent.range, thing.range)) {
                    parent.appendChild(thing);
                    stack.push(parent);
                    stack.push(thing);
                    break;
                }
                if (parent === root) {
                    // impossible ?!
                    break;
                }
            }
        }
        // remove helper usage-nodes
        stack.length = 0;
        stack.push(root);
        while (stack.length > 0) {
            let n = stack.pop();
            if (n instanceof Usage && n.isHelper) {
                n.remove();
            }
            else {
                stack.push(...n.children());
            }
        }
    }
    static _compareCaptures(a, b) {
        return a.node.startIndex - b.node.startIndex;
    }
    static _compareByRange(a, b) {
        return common_1.compareRangeByStart(a.range, b.range);
    }
    debugPrint() {
        console.log(this.root.toString());
    }
}
exports.Locals = Locals;
class Node {
    constructor(range, type) {
        this.range = range;
        this.type = type;
        this._children = [];
    }
    children() {
        return this._children;
    }
    remove() {
        if (!this._parent) {
            return false;
        }
        const idx = this._parent._children.indexOf(this);
        if (idx < 0) {
            return false;
        }
        this._parent._children.splice(idx, 1);
        return true;
    }
    appendChild(node) {
        this._children.push(node);
        node._parent = this;
    }
    toString() {
        return `${this.type}@${this.range.start.line},${this.range.start.character}-${this.range.end.line},${this.range.end.character}`;
    }
}
class Usage extends Node {
    constructor(name, range, isHelper) {
        super(range, 2 /* Usage */);
        this.name = name;
        this.range = range;
        this.isHelper = isHelper;
    }
    appendChild(_node) {
        // console.log('ignored', node.toString());
    }
    toString() {
        return `use:${this.name}`;
    }
}
exports.Usage = Usage;
class Definition extends Node {
    constructor(name, range) {
        super(range, 1 /* Definition */);
        this.name = name;
        this.range = range;
    }
    appendChild(_node) {
        // console.log('ignored', node.toString());
    }
    toString() {
        return `def:${this.name}`;
    }
}
exports.Definition = Definition;
class Scope extends Node {
    constructor(range, likelyExports) {
        super(range, 0 /* Scope */);
        this.likelyExports = likelyExports;
    }
    *definitions() {
        for (let item of this._children) {
            if (item instanceof Definition) {
                yield item;
            }
        }
    }
    *usages() {
        for (let item of this._children) {
            if (item instanceof Usage) {
                yield item;
            }
        }
    }
    *scopes() {
        for (let item of this._children) {
            if (item instanceof Scope) {
                yield item;
            }
        }
    }
    findScope(position) {
        for (let scope of this.scopes()) {
            if (common_1.containsPosition(scope.range, position)) {
                return scope.findScope(position);
            }
        }
        return this;
    }
    findDefinitionOrUsage(position) {
        for (let child of this._children) {
            if ((child instanceof Definition || child instanceof Usage) && common_1.containsPosition(child.range, position)) {
                return child;
            }
        }
    }
    findDefinitions(text) {
        const result = [];
        for (let child of this.definitions()) {
            if (child.name === text) {
                result.push(child);
            }
        }
        if (result.length > 0) {
            return result;
        }
        if (!(this._parent instanceof Scope)) {
            return [];
        }
        return this._parent.findDefinitions(text);
    }
    findUsages(text) {
        const bucket = [];
        // find higest scope defining
        let scope = this;
        while (!scope._defines(text)) {
            if (scope._parent instanceof Scope) {
                scope = scope._parent;
            }
            else {
                break;
            }
        }
        // find usages in all child scope (unless also defined there)
        scope._findUsagesDown(text, bucket);
        return bucket.flat();
    }
    _findUsagesDown(text, bucket) {
        // usages in this scope
        const result = [];
        for (let child of this.usages()) {
            if (child.name === text) {
                result.push(child);
            }
        }
        bucket.push(result);
        // usages in child scope (unless also defined there)
        for (let child of this.scopes()) {
            if (!child._defines(text)) {
                child._findUsagesDown(text, bucket);
            }
        }
    }
    _defines(text) {
        for (let child of this.definitions()) {
            if (child.name === text) {
                return true;
            }
        }
        return false;
    }
    toString(depth = 0) {
        let scopes = [];
        let parts = [];
        this._children.slice(0).forEach(child => {
            if (child instanceof Scope) {
                scopes.push(child.toString(depth + 2));
            }
            else {
                parts.push(child.toString());
            }
        });
        let indent = ' '.repeat(depth);
        let res = `${indent}Scope@${this.range.start.line},${this.range.start.character}-${this.range.end.line},${this.range.end.character}`;
        res += `\n${indent + indent}${parts.join(`, `)}`;
        res += `\n${indent}${scopes.join(`\n${indent}`)}`;
        return res;
    }
}
exports.Scope = Scope;
