"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SymbolIndex = void 0;
const lsp = __importStar(require("vscode-languageserver"));
const common_1 = require("../common");
const trie_1 = require("../util/trie");
const documentSymbols_1 = require("./documentSymbols");
const languages_1 = __importDefault(require("../languages"));
class Queue {
    constructor() {
        this._queue = new Set();
    }
    enqueue(uri) {
        if (common_1.isInteresting(uri) && !this._queue.has(uri)) {
            this._queue.add(uri);
        }
    }
    dequeue(uri) {
        this._queue.delete(uri);
    }
    consume() {
        const result = Array.from(this._queue.values());
        this._queue.clear();
        return result;
    }
}
class Cache {
    constructor() {
        this._definitions = trie_1.Trie.create();
        this._usages = trie_1.Trie.create();
        this._cleanup = new Map();
    }
    get definitions() {
        return this._definitions;
    }
    get usages() {
        return this._usages;
    }
    insertDefinition(text, definition) {
        let all = this._definitions.get(text);
        if (all) {
            all.add(definition);
        }
        else {
            all = new Set([definition]);
            this._definitions.set(text, all);
        }
        this._addCleanup(definition.location.uri, () => {
            if (all.delete(definition) && all.size === 0) {
                this._definitions.delete(text);
            }
        });
    }
    insertUsage(text, usage) {
        let all = this._usages.get(text);
        if (all) {
            all.add(usage);
        }
        else {
            all = new Set([usage]);
            this._usages.set(text, all);
        }
        this._addCleanup(usage.uri, () => {
            if (all.delete(usage) && all.size === 0) {
                this._usages.delete(text);
            }
        });
    }
    _addCleanup(uri, cleanupFn) {
        const arr = this._cleanup.get(uri);
        if (arr) {
            arr.push(cleanupFn);
        }
        else {
            this._cleanup.set(uri, [cleanupFn]);
        }
    }
    delete(uri) {
        const callbacks = this._cleanup.get(uri);
        if (callbacks) {
            callbacks.forEach(fn => fn());
            this._cleanup.delete(uri);
            return true;
        }
        return false;
    }
    toString() {
        return `${this._definitions.size} symbols, ${this._usages.size} usages`;
    }
}
class SymbolIndex {
    constructor(_trees, _documents) {
        this._trees = _trees;
        this._documents = _documents;
        this._cache = new Cache();
        this._queue = new Queue();
    }
    get definitions() {
        return this._cache.definitions;
    }
    get usages() {
        return this._cache.usages;
    }
    addFile(uris) {
        if (Array.isArray(uris)) {
            uris.forEach(this._queue.enqueue, this._queue);
        }
        else {
            this._queue.enqueue(uris);
        }
    }
    removeFile(uris) {
        for (let uri of uris) {
            this._queue.dequeue(uri);
            this._cache.delete(uri);
        }
    }
    async update() {
        await this._currentUpdate;
        this._currentUpdate = this._doUpdate();
        return this._currentUpdate;
    }
    async _doUpdate() {
        const uris = this._queue.consume();
        if (uris.length !== 0) {
            // clear cached info for changed uris
            const sw = new common_1.StopWatch();
            const delCount = uris.map(this._cache.delete, this._cache).filter(Boolean).length;
            console.log(`[index] removed ${delCount} files ${sw.elapsed()}ms`);
            // schedule a new task to update the cache for`
            // changed uris
            sw.reset();
            const tasks = uris.map(this._createIndexTask, this);
            const stats = await common_1.parallel(tasks, 50, new lsp.CancellationTokenSource().token);
            let totalRetrieve = 0;
            let totalIndex = 0;
            for (let stat of stats) {
                totalRetrieve += stat.durationRetrieve;
                totalIndex += stat.durationIndex;
            }
            console.log(`[index] added ${uris.length} files ${sw.elapsed()}ms, retrieval: ${Math.round(totalRetrieve)}ms, index: ${Math.round(totalIndex)}ms`);
            console.log(`[usage] info: ${this._cache.toString()}`);
        }
    }
    _createIndexTask(uri) {
        return async () => {
            const _t1Retrieve = performance.now();
            const document = await this._documents.retrieve(uri);
            const durationRetrieve = performance.now() - _t1Retrieve;
            const _t1Index = performance.now();
            try {
                await this._doIndex(document);
            }
            catch (e) {
                console.log(`FAILED to index ${uri}`, e);
            }
            const durationIndex = performance.now() - _t1Index;
            return { durationRetrieve, durationIndex };
        };
    }
    async _doIndex(document) {
        // (1) use outline information to feed the global index of definitions
        const symbols = documentSymbols_1.Outline.create(document, this._trees);
        const walkSymbols = (symbols, parent) => {
            for (let symbol of symbols) {
                const info = lsp.SymbolInformation.create(symbol.name, symbol.kind, symbol.selectionRange, document.uri, parent?.name);
                if (symbol.children) {
                    walkSymbols(symbol.children, symbol);
                }
                this._cache.insertDefinition(info.name, info);
            }
        };
        walkSymbols(symbols, undefined);
        // (2) Use usage-queries to feed the global index of usages.
        const tree = this._trees.getParseTree(document);
        if (tree) {
            const query = languages_1.default.getQuery(document.languageId, 'references');
            const captures = query.captures(tree.rootNode);
            for (let capture of captures) {
                const usage = lsp.Location.create(document.uri, common_1.asLspRange(capture.node));
                this._cache.insertUsage(capture.node.text, usage);
            }
        }
    }
}
exports.SymbolIndex = SymbolIndex;
