"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.Validation = void 0;
const vscode_languageserver_1 = require("vscode-languageserver");
const common_1 = require("../common");
class Validation {
    constructor(_connection, documents, _trees) {
        this._connection = _connection;
        this._trees = _trees;
        this._currentValidation = new Map();
        documents.all().forEach(this._triggerValidation, this);
        documents.onDidChangeContent(e => this._triggerValidation(e.document));
        documents.onDidOpen(e => this._triggerValidation(e.document));
        documents.onDidClose(e => {
            _connection.sendDiagnostics({ uri: e.document.uri, diagnostics: [] });
        });
    }
    async _triggerValidation(document) {
        if (!common_1.isInteresting(document.uri)) {
            // unsupported
            return;
        }
        const config = await this._connection.workspace.getConfiguration({ section: 'anycode', scopeUri: document.uri });
        if (!config.diagnostics) {
            return;
        }
        // cancel pending validation
        let cts = this._currentValidation.get(document);
        cts?.cancel();
        cts?.dispose();
        // schedule new validation
        cts = new vscode_languageserver_1.CancellationTokenSource();
        this._currentValidation.set(document, cts);
        const handle = setTimeout(() => this._createDiagnostics(document), 500);
        cts.token.onCancellationRequested(() => clearTimeout(handle));
    }
    async _createDiagnostics(document) {
        const tree = this._trees.getParseTree(document);
        const diagnostics = [];
        if (tree) {
            // find MISSING nodes (those that got auto-inserted)
            const cursor = tree.walk();
            const seen = new Set();
            try {
                let visitedChildren = false;
                while (true) {
                    if (cursor.nodeIsMissing && !seen.has(cursor.nodeId)) {
                        diagnostics.push({
                            range: common_1.asLspRange(cursor.currentNode()),
                            message: `Expected '${cursor.nodeType}'`,
                            severity: vscode_languageserver_1.DiagnosticSeverity.Error,
                            source: 'anycode',
                            code: 'missing',
                        });
                        seen.add(cursor.nodeId);
                    }
                    // depth first search
                    if (!visitedChildren) {
                        if (!cursor.gotoFirstChild()) {
                            visitedChildren = true;
                        }
                    }
                    if (visitedChildren) {
                        if (cursor.gotoNextSibling()) {
                            visitedChildren = false;
                        }
                        else if (cursor.gotoParent()) {
                            visitedChildren = true;
                        }
                        else {
                            break;
                        }
                    }
                }
            }
            finally {
                cursor.delete();
            }
            // // find "generic" error nodes
            // let query: Parser.Query | undefined;
            // try {
            // 	query = tree.getLanguage().query('(ERROR) @error');
            // 	const captures = query.captures(tree.rootNode);
            // 	for (let capture of captures) {
            // 		diags.push({
            // 			range: asCodeRange(capture.node),
            // 			message: 'Error',
            // 			severity: vscode.DiagnosticSeverity.Error,
            // 			source: 'anycode',
            // 			code: 'error',
            // 		});
            // 	}
            // } catch {
            // 	// ignore - parsing the query might fail
            // } finally {
            // 	query?.delete();
            // }
        }
        this._connection.sendDiagnostics({ uri: document.uri, diagnostics });
    }
}
exports.Validation = Validation;
