"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const tree_sitter_1 = __importDefault(require("../tree-sitter/tree-sitter"));
const c_sharp_1 = __importDefault(require("./queries/c_sharp"));
const c_1 = __importDefault(require("./queries/c"));
const cpp_1 = __importDefault(require("./queries/cpp"));
const go_1 = __importDefault(require("./queries/go"));
const java_1 = __importDefault(require("./queries/java"));
const php_1 = __importDefault(require("./queries/php"));
const python_1 = __importDefault(require("./queries/python"));
const rust_1 = __importDefault(require("./queries/rust"));
const typescript_1 = __importDefault(require("./queries/typescript"));
const _queryModules = new Map([
    ['csharp', c_sharp_1.default],
    ['c', c_1.default],
    ['cpp', cpp_1.default],
    ['go', go_1.default],
    ['java', java_1.default],
    ['php', php_1.default],
    ['python', python_1.default],
    ['rust', rust_1.default],
    ['typescript', typescript_1.default],
]);
class Languages {
    static async init(langConfiguration) {
        this._langConfiguration = langConfiguration;
        for (const [entry, config] of langConfiguration) {
            const lang = await tree_sitter_1.default.Language.load(entry.wasmUri);
            this._languageInstances.set(entry.languageId, lang);
            this._configurations.set(entry.languageId, config);
        }
    }
    static getLanguage(languageId) {
        let result = this._languageInstances.get(languageId);
        if (!result) {
            console.warn(`UNKNOWN languages: '${languageId}'`);
            return undefined;
        }
        return result;
    }
    static allAsSelector() {
        return [...this._languageInstances.keys()];
    }
    static getQuery(languageId, type, strict = false) {
        const module = _queryModules.get(languageId);
        if (!module) {
            // unknown language or invalid query (deleted after failed parse attempt)
            return this.getLanguage(languageId).query('');
        }
        const source = module[type] ?? '';
        const key = `${languageId}/${type}`;
        let query = this._queryInstances.get(key);
        if (!query) {
            try {
                query = this.getLanguage(languageId).query(source);
            }
            catch (e) {
                query = this.getLanguage(languageId).query('');
                console.error(languageId, e);
                if (strict) {
                    throw e;
                }
            }
            this._queryInstances.set(key, query);
        }
        return query;
    }
    static getSupportedLanguages(feature, types) {
        const result = [];
        for (let languageId of this._languageInstances.keys()) { // USE actually supported languages
            const module = _queryModules.get(languageId);
            if (!module) {
                console.warn(`${languageId} NOT supported by queries`);
                continue;
            }
            for (let type of types) {
                if (module[type] && this._configurations.get(languageId)?.[feature]) {
                    result.push(languageId);
                    break;
                }
            }
        }
        return result;
    }
    static getLanguageIdByUri(uri) {
        let end = uri.lastIndexOf('?');
        if (end < 0) {
            end = uri.lastIndexOf('#');
        }
        if (end > 0) {
            uri = uri.substring(0, end);
        }
        const start = uri.lastIndexOf('.');
        const suffix = uri.substring(start + 1);
        for (let [info] of this._langConfiguration) {
            for (let candidate of info.suffixes) {
                if (candidate === suffix) {
                    return info.languageId;
                }
            }
        }
        return `unknown/${uri}`;
    }
}
exports.default = Languages;
Languages._languageInstances = new Map();
Languages._queryInstances = new Map();
Languages._configurations = new Map();
