"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const node_1 = require("vscode-languageserver/node");
const vscode_languageserver_1 = require("vscode-languageserver");
const tree_sitter_1 = __importDefault(require("../tree-sitter/tree-sitter"));
const trees_1 = require("./trees");
const documentSymbols_1 = require("./features/documentSymbols");
const symbolIndex_1 = require("./features/symbolIndex");
const selectionRanges_1 = require("./features/selectionRanges");
const completions_1 = require("./features/completions");
const workspaceSymbols_1 = require("./features/workspaceSymbols");
const definitions_1 = require("./features/definitions");
const references_1 = require("./features/references");
const validation_1 = require("./features/validation");
const documentStore_1 = require("./documentStore");
const languages_1 = __importDefault(require("./languages"));
const foldingRanges_1 = require("./features/foldingRanges");
const documentHighlights_1 = require("./features/documentHighlights");
const connection = node_1.createConnection();
// patch console.log/warn/error calls
console.log = connection.console.log.bind(connection.console);
console.warn = connection.console.warn.bind(connection.console);
console.error = connection.console.error.bind(connection.console);
const features = [];
connection.onInitialize(async (params) => {
    const initData = params.initializationOptions;
    // (1) init tree sitter and languages before doing anything else
    const options = {
        locateFile() {
            return initData.treeSitterWasmUri;
        }
    };
    await tree_sitter_1.default.init(options);
    await languages_1.default.init(initData.supportedLanguages);
    // (2) setup features
    const documents = new documentStore_1.DocumentStore(connection);
    const trees = new trees_1.Trees(documents);
    const symbolIndex = new symbolIndex_1.SymbolIndex(trees, documents);
    features.push(new workspaceSymbols_1.WorkspaceSymbol(symbolIndex));
    features.push(new documentSymbols_1.DocumentSymbols(documents, trees));
    features.push(new definitions_1.DefinitionProvider(documents, trees, symbolIndex));
    features.push(new references_1.ReferencesProvider(documents, trees, symbolIndex));
    features.push(new documentHighlights_1.DocumentHighlightsProvider(documents, trees));
    features.push(new completions_1.CompletionItemProvider(documents, trees, symbolIndex));
    features.push(new selectionRanges_1.SelectionRangesProvider(documents, trees));
    features.push(new foldingRanges_1.FoldingRangeProvider(documents, trees));
    new validation_1.Validation(connection, documents, trees);
    // (nth) manage symbol index. add/remove files as they are disovered and edited
    documents.all().forEach(doc => symbolIndex.addFile(doc.uri));
    documents.onDidOpen(event => symbolIndex.addFile(event.document.uri));
    documents.onDidChangeContent(event => symbolIndex.addFile(event.document.uri));
    connection.onNotification('queue/remove', uris => symbolIndex.removeFile(uris));
    connection.onNotification('queue/add', uris => symbolIndex.addFile(uris));
    connection.onRequest('queue/init', uris => {
        symbolIndex.addFile(uris);
        return symbolIndex.update();
    });
    console.log('Tree-sitter, languages, and features are READY');
    return {
        capabilities: { textDocumentSync: vscode_languageserver_1.TextDocumentSyncKind.Incremental }
    };
});
connection.onInitialized(async () => {
    for (let feature of features) {
        feature.register(connection);
    }
});
connection.listen();
