"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Trees = void 0;
const lruMap_1 = require("./util/lruMap");
const tree_sitter_1 = __importDefault(require("../tree-sitter/tree-sitter"));
const languages_1 = __importDefault(require("./languages"));
class Entry {
    constructor(version, tree, edits) {
        this.version = version;
        this.tree = tree;
        this.edits = edits;
    }
}
class Trees {
    constructor(_documents) {
        this._documents = _documents;
        this._cache = new lruMap_1.LRUMap({
            size: 100,
            dispose(entries) {
                for (let [, value] of entries) {
                    value.tree.delete();
                }
            }
        });
        this._listener = [];
        this._parser = new tree_sitter_1.default();
        // build edits when document changes
        this._listener.push(_documents.onDidChangeContent2(e => {
            const info = this._cache.get(e.document.uri);
            if (info) {
                info.edits.push(Trees._asEdits(e));
            }
        }));
    }
    dispose() {
        this._parser.delete();
        for (let item of this._cache.values()) {
            item.tree.delete();
        }
        for (let item of this._listener) {
            item.dispose();
        }
    }
    getParseTree(documentOrUri) {
        if (typeof documentOrUri === 'string') {
            return this._documents.retrieve(documentOrUri).then(doc => this._parse(doc));
        }
        else {
            return this._parse(documentOrUri);
        }
    }
    _parse(documentOrUri) {
        let info = this._cache.get(documentOrUri.uri);
        if (info?.version === documentOrUri.version) {
            return info.tree;
        }
        const language = languages_1.default.getLanguage(documentOrUri.languageId);
        if (!language) {
            return undefined;
        }
        this._parser.setLanguage(language);
        this._parser.setTimeoutMicros(1000 * 1000); // parse max 1sec
        try {
            const version = documentOrUri.version;
            const text = documentOrUri.getText();
            if (!info) {
                // never seen before, parse fresh
                const tree = this._parser.parse(text);
                info = new Entry(version, tree, []);
                this._cache.set(documentOrUri.uri, info);
            }
            else {
                // existing entry, apply deltas and parse incremental
                const oldTree = info.tree;
                const deltas = info.edits.flat();
                deltas.forEach(delta => oldTree.edit(delta));
                info.edits.length = 0;
                info.tree = this._parser.parse(text, oldTree);
                info.version = version;
                oldTree.delete();
            }
            return info.tree;
        }
        catch (e) {
            this._cache.delete(documentOrUri.uri);
            return undefined;
        }
    }
    static _asEdits(event) {
        return event.changes.map(change => ({
            startPosition: this._asTsPoint(change.range.start),
            oldEndPosition: this._asTsPoint(change.range.end),
            newEndPosition: this._asTsPoint(event.document.positionAt(change.rangeOffset + change.text.length)),
            startIndex: change.rangeOffset,
            oldEndIndex: change.rangeOffset + change.rangeLength,
            newEndIndex: change.rangeOffset + change.text.length
        }));
    }
    static _asTsPoint(position) {
        const { line: row, character: column } = position;
        return { row, column };
    }
}
exports.Trees = Trees;
;
