"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.Trie = void 0;
class Entry {
    constructor(key, value) {
        this.key = key;
        this.value = value;
    }
}
class Trie {
    constructor(ch, element) {
        this.ch = ch;
        this.element = element;
        this._size = 0;
        this._depth = 0;
        this._children = new Map();
    }
    static create() {
        return new Trie('', undefined);
    }
    get size() {
        return this._size;
    }
    get depth() {
        return this._depth;
    }
    set(str, element) {
        let chars = Array.from(str);
        let node = this;
        for (let pos = 0; pos < chars.length; pos++) {
            node._depth = Math.max(chars.length - pos, node._depth);
            const ch = chars[pos];
            let child = node._children.get(ch);
            if (!child) {
                child = new Trie(ch, undefined);
                node._children.set(ch, child);
            }
            node = child;
        }
        if (!node.element) {
            this._size += 1;
            node.element = new Entry(str, element);
        }
        else {
            node.element.value = element;
        }
    }
    get(str) {
        let chars = Array.from(str);
        let node = this;
        for (let pos = 0; pos < chars.length; pos++) {
            const ch = chars[pos];
            let child = node._children.get(ch);
            if (!child) {
                return undefined;
            }
            node = child;
        }
        return node.element?.value;
    }
    delete(str) {
        let chars = Array.from(str);
        let node = this;
        let path = [];
        for (let pos = 0; pos < chars.length; pos++) {
            const ch = chars[pos];
            let child = node._children.get(ch);
            if (!child) {
                return false;
            }
            path.push([ch, node]);
            node = child;
        }
        if (!node.element) {
            return false;
        }
        // unset element
        node.element = undefined;
        this._size -= 1;
        // cleanup parents and update depths
        while (path.length > 0) {
            // parent
            const [nodeCh, parent] = path.pop();
            if (node._children.size === 0 && !node.element) {
                parent._children.delete(nodeCh);
            }
            node = parent;
            if (node._children.size === 0) {
                node._depth = 0;
            }
            else {
                let newDepth = 0;
                for (let child of node._children.values()) {
                    newDepth = Math.max(newDepth, child.depth);
                }
                node._depth = 1 + newDepth;
            }
        }
        return true;
    }
    *query(str) {
        let bucket = new Set();
        this._query(str, 0, bucket);
        for (let item of bucket) {
            yield* item;
        }
    }
    _query(str, pos, bucket) {
        if (bucket.has(this)) {
            return;
        }
        if (pos >= str.length) {
            bucket.add(this);
            return;
        }
        if (str.length - pos > this._depth) {
            // there is more characters left than there are nodes
            return;
        }
        for (let [ch, child] of this._children) {
            if (ch.toLowerCase() === str[pos].toLowerCase()) {
                child._query(str, pos + 1, bucket);
            }
            if (pos > 0) {
                // only proceed fuzzy if the first character has matched
                child._query(str, pos, bucket);
            }
        }
    }
    *_entries() {
        if (this.element) {
            yield [this.element.key, this.element.value];
        }
        for (let child of this._children.values()) {
            yield* child._entries();
        }
    }
    [Symbol.iterator]() {
        return this._entries();
    }
}
exports.Trie = Trie;
